/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (collect.C)
  
  Copyright:  Damian J. J. Farnell, Mar2000

  Updated:    Joerg Schulenburg 2004-2008

  ===================================================

  ---------------------------------------------------

  PROGRAM NAME: collect.C 

  PURPOSE: collect same configurations to reduce loops

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>
// #include <math.h>
 #include <string.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"
// #include "nr.h"
// #include "nrutil.h"
#include "storage.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

#ifndef DO_COLLECT
#define DO_COLLECT 1  /* set to 0 to switch off collect, default = 1 */
#endif

/* Joerg05:
 * who is calling  collect_term?
 *  62 times by obtain()+term2()..term62() in obtain.C 
 * parameters and its ranges?
 *  indices[0] - configuration (lattice animal), 1..NumberConfigurations
 *                                          new: 0 for GS Energy
 *  indices[1] - term 2..62, STOP_TERM=0=end (not used in solve!)
 *  indices[2..5] -  0,1..NumConfig ??  (0,0,0,0)=not_used(check it)
 * ToDo: exchange [0] <-> [1]
 * Outputfile fp1=*.terms is used for solve()
 * Outputfile fp1=*_gs    is used for ...     (new Sep05) ???
 * maximum number of terms and clusters for sample models?
 *  MAX_NUMBER_TERMS_IN_CCM_EQNS to big? (*46Bytes)
 * ToDo: dynamic arrays
 */

// MAX_NUMBER_TERMS_IN_CCM_EQNS renamed to MAXNUM_HASHED_TERMS
// MAXNUM_HASHED_TERMS should be a prime number for good hashing
// static arrays: MAXNUM_HASHED_TERMS * 44 byte, dfl=(1<<20)*44B=44MB

const int MAXNUM_HASHED_TERMS=(1<<20)-3; // 2^19-1 for s=1/2-LC-14  * 44 byte

// ToDo: move this table to obtain.C and ...
// ToDo: add termGSE or use term13 + debugstring like below
//  ...  use as debug arguments (...,1(="zz"),"function")
//   this table is wrong for id=3 but valid for id=1,2
//  xxzLC, xxzSQ: SzSz=1.2..1.5 + SxSy+SySx=8.26+9.27..9.45
const char *term62s[62+2]={"",         // 0.0  ToDo: reduce 62 by 10 k_to_m's?
//     term13[3]      term62[20] numcoeffs[5] 
       "zz                  ??? ?[0]", // 1.1 not used ???
       "zz           Fm.S+m.Szk  [1]", // 1.2        xxzLC(z) isingLC
       "zz           Fk.S+k.Szm  [1]", // 1.3 = ~1.2 ...
       "zz          Hkm.S+k.S+m  [1]", // 1.4        ...
       "zz        Fk.Fm.S+k.S+m  [2]", // 1.5        xxzLC(z) isingLC
       "z+              S+m.Szk  [0]", // 2.6        kago_triangle? lc_mf? MB
       "z+          Fkl.S+k.S+m  [1]", // 2.7
       "z-       -2*Fml.Szk.Szm  [1]", // 3.8 cancels 5.17? kago_triangle?
       "z-      -2*Hkml.S+k.Szm  [1]",
       "z-       -Fm.Fm.S+m.Szk  [2]",
       "z-    -2*Hkm.Fm.S+k.S+m  [2]",
       "z-     -2*Fk.Fm.S+k.Szm  [2]",
       "z-    -Fk.Fm.Fm.S+k.S+m  [3]",
       "+z              S+k.Szm  [0]", // 4.14 = ~2.6       kago_triangle?
       "+z           Fm.S+m.S+k  [1]", // 4.15 = ~2.7 Fml ???
       "-z      -2*Hmkl.S+m.Szk  [1]", // 5.16              kago_triangle?
       "-z       -2*Fml*Szk*Szm  [1]", // 5.17 cancels 3.8?
       "-z     -2*Fm.Fk.S+m.Szk  [2]",
       "-z    -2*Hkm.Fk.S+k.S+m  [2]",
       "-z           -Fm.Fm.S+m  [2]",
       "-z    -Fk.Fk.Fm.S+k.S+m  [3]",
       "+-        -2*Fm.S+k.Szm  [1]", // 6.22 CaVO_neel? j1j2_aus_srcubo kago
       "+-       -Fm.Fm.S+k.S+m  [2]",
       "-+        -2*Fk.S+m.Szk  [1]", // 7.24 CaVO_neel?
       "-+       -Fk.Fk.S+k.S+m  [2]",
       "++              S+k.S+m  [0]", // 8.26  xxzLC(xy)
       "--        4*Hkm.Szk.Szm  [1]", // 9.27  xxzLC(xy)
       "--    2*Hkm.Hkm.S+k.S+m  [2]", // 9.28  ...
       "--     4*Hkm.Fm.S+m.Szk  [2]", // 9.29
       "--      4*Fk.Fm.Szk.Szm  [2]", // 9.30
       "--   2*Fm.Fm.Fk.S+m.Szk  [3]",
       "--  4*Fk.Fm.Hkm.S+k.S+m  [3]",
       "--   2*Fk.Fk.Fm.S+k.Szm  [3]",
       "--  Fk.Fk.Fm.Fm.S+k.S+m  [4]", // 9.34  (biggest contribution CPUh)
       "--     4*Hkm.Fk.S+k.Szm  [2]", // 9.35 = ~9.29
       "--       2*Mkkm.S+k.Szm  [1] 2s>1", // 9.36
       "--       2*Mkmm.S+m.Szz  [1] NOTE: DIRECTIONALITY TO BE ADDED FOR TRIANGLE!",
       "--        Nkkmm.S+k.S+m  [1]",
       "--      Fm.Mkkm.S+k.S+m  [2]", // 9.39
       "--     2*Fk.Hmm.S+m.Szk  [2]", // 9.40
       "--      Fk.Mmmk.S+k.S+m  [2]", // 9.41 = ~9.39
       "--     2*Fm.Hkk.S+k.Szm  [2]", // 9.42 = ~9.40
       "--      Hkk.Hmm.S+k.S+m  [2]",
       "--    Hkk.Fm.Fm.S+k.S+m  [3] 2s>1", //  9.44         ...
       "--    Hmm.Fk.Fk.S+k.S+m  [3] 2s>1", //  9.45 = ~9.44 xxzLC(xy)
       "z-         -Hmm.S+m.Szk  [1] 2s>1", //  3.46
       "z-        -Mkmm.S+k.S+m  [1] 2s>1", //  3.47
       "z-      -Fk.Hmm.S+k.S+m  [2] 2s>1", //  3.48
       "-z        -Mkkm.S+k.S+m  [1] 2s>1", //  5.49 = ~3.47
       "-z         -Hkk.S+k.Szm  [1] 2s>1", //  5.50 = ~3.46
       "-z      -Fm.Hkk.S+k.S+m  [2] 2s>1", //  5.51 = ~3.48
       "+-         -Hmm.S+k.S+m  [1] 2s>1", //  6.52  ToDo: sample system?
       "-+         -Hkk.S+k.S+m  [1] 2s>1", //  7.53 = ~6.52
       "z                Fk.S+k  [1]",      // 10.54  lc_mf
       "-             -2*Fk.Szk  [1]",      // 11.55  lc_mf
       "-              -Hkk.S+k  [1]",
       "-            -Fk.Fk.S+k  [2]",
       "z^2         2Fk.S+k.Szk  [1]",      // 12.58
       "z^2          Fk.S+k.S+k  [1]",      // 12.59
       "z^2         Hkk.S+k.S+k  [1]",
       "z^2       Fk.Fk.S+k.S+k  [2]",
       "+                   S+k  [0]",      // 13.62  lc_mf
       ""};
const char *term81s[81+2]={"",         // 0.0
//     term13[3] term81[20] numcoeffs[5] 
       "zz           Pi.S+i.Szj  [1]", // 1.1
       "zz        Pi.Fj.S+i.S+j  [2]",
       "zz           Pj.S+j.Szi  [1]",
       "zz        Pj.Fi.S+i.S+j  [2]",
       "zz          Qij.S+i.S+j  [1]",
       "z+           Pi.S+i.S+m  [1]", // 2.6
       "z-     -2*Pi.Fj.S+i.Szj  [2]", // 3.7
       "z-      -Pi.Gjj.S+i.S+j  [2]", // 3.8
       "z-    -Pi.Fj.Fj.S+i.S+j  [3]", // 3.9
       "z-         -2Pj.Szi.Szj  [1]", // 3.10
       "z-     -2*Pj.Fj.S+j.Szi  [2]", // 3.11
       "z-     -2*Pj.Fi.S+i.Szj  [2]", // 3.12
       "z-    -2*Pj.Gij.S+i.S+j  [2]", // 3.13
       "z-  -2*Pj.Fi.Fj.S+i.S+j  [3]", // 3.14
       "z-       -2*Gij.S+i.Szj  [1]", // 3.15
       "z-    -2*Qij.Fj.S+i.S+j  [2]", // 3.16
       "z-         -Qjj.S+j.Szi  [1]", // 3.17
       "z-      -Qjj.Fi.S+i.S+j  [2]", // 3.18
       "z-        -Lijj.S+i.S+j  [1]", // 3.19
       "+z           Pj.S+i.S+j  [1]", // 4.20 like 2.6 backwards
       "-z     -2*Pj.Fi.S+j.Szi  [2]", // 5.21 like 3.7 backwards
       "-z      -Pj.Gii.S+i.S+j  [2]", // 5.22 like 3.8 backwards
       "-z    -Pj.Fi.Fi.S+i.S+j  [3]", // 5.23 like 3.9 backwards
       "-z        -2*Pi.Szi.Szj  [1]", // 5.24 like 3.10 backwards
       "-z     -2*Pi.Fj.S+j.Szi  [2]", // 5.25 like 3.12 backwards
       "-z     -2*Pi.Fi.S+i.Szj  [2]", // 5.26 like 3.11 backwards
       "-z    -2*Pi.Gij.S+i.S+j  [2]", // 5.27 like 3.13 backwards
       "-z  -2*Pi.Fi.Fj.S+i.S+j  [3]", // 5.28 like 3.14 backwards
       "-z       -2*Qij.S+j.Szi  [1]", // 5.29 ???
       "-z    -2*Qij.Fi.S+i.S+j  [2]", // 5.30 like 3.16
       "-z         -Qii.S+i.Szj  [1]", // 5.31 like 3.17
       "-z    -2*Qii.Fj.S+i.S+j  [2]", // 5.32 like 3.18
       "-z        -Liij.S+i.S+j  [1]", // 5.33 like 3.19
       "+-        -2*Pj.S+i.Szj  [1]", // 6.34
       "+-     -2*Pj.Fj.S+i.S+j  [2]", // 6.35
       "+-         -Qjj.S+i.S+j  [1]", // 6.36
       "-+        -2*Pi.S+j.Szi  [1]", // 7.37 like 6.34
       "-+     -2*Pi.Fi.S+i.S+j  [2]", // 7.38 like 6.35
       "-+         -Qii.S+i.S+j  [1]", // 7.39 like 6.36
       "--      4*Pi.Fj.Szi.Szj  [2]", // 9.40
       "--     4*Pi.Gij.S+i.Szj  [2]", // 9.41 like 9.62 but forward
       "--     2*Pi.Gjj.S+j.Szi  [2]",
       "--    2*Pi.Hijj.S+i.S+j  [2]",
       "--   4*Pk.Fk.Fm.S+k.Szm  [3]",
       "--  4*Pk.Fm.Hkm.S+k.S+m  [3]",
       "--  2*Pi.Fi.Gjj.S+i.S+j  [3]",
       "-- 2Pk.Fk.Fm.Fm.S+k.S+m  [4]", // 9.47 same as 9.69 but forward
       "--   2*Pk.Fm.Fm.S+m.Szj  [3]", // 9.48
       "--     2*Qkk.Fj.S+i.Szj  [2]",
       "--      Qkk.Hmm.S+k.S+m  [2]",
       "--    Qkk.Fm.Fm.S+k.S+m  [3]",
       "--        4*Qkm.Szk.Szm  [1]", // 9.52
       "--     4*Qkm.Fm.S+m.Szk  [2]", // 9.53
       "--     4*Qij.Fi.S+i.Szj  [2]", // 9.54
       "--    4*Qkm.Hkm.S+k.S+m  [2]", // 9.55
       "--  4*Qkm.Fk.Fm.S+k.S+m  [3]", // 9.56
       "--       2*Lkkm.S+k.Szm  [1]",
       "--    2*Lkkm.Fm.S+k.S+m  [2]",
       "--       2*Lijj.S+j.Szi  [1]", // 9.59 NOTE: DIRECTIONALITY TO BE ADDED FOR TRIANGLE!
       "--    2*Lmmk.Fk.S+k.S+m  [2]",
       "--        Mkkmm.S+k.S+m  [1]",
       "--     4*Pj.Gij.S+j.Szi  [2]", // 9.62 like 9.41 but backwards
       "--    2*Pi.Lijj.S+i.S+j  [2]",
       "--      4*Pj.Fi.Szi.Szj  [2]", // most time consuming?
       "--   4*Pj.Fi.Fj.S+j.Szi  [3]",
       "--     2*Pj.Gii.S+i.Szj  [2]",
       "--  2*Pj.Fj.Gii.S+i.S+j  [3]",
       "--   2*Pm.Fk.Fk.S+k.Szm  [3]",
       "-- 2Pm.Fm.Fk.Fk.S+k.S+m  [4]", // 9.69 same as 9.47 but k <=> m
       "--  4*Pj.Fi.Gkm.S+k.S+m  [3]",
       "--     2*Qjj.Fi.S+j.Szi  [2]",
       "--      Qjj.Hii.S+i.S+j  [2]",
       "--    Qjj.Fi.Fi.S+k.S+m  [3]",
       "z                Pk.S+k  [1]", // 10.74
       "-             -2*Pk.Szk  [1]", // 11.75
       "-          -2*Pk.Fk.S+k  [2]",
       "-              -Qkk.S+k  [1]",
       "z^2        2*Pk.S+k.Szk  [1]", // 12.78
       "z^2          Pk.S+k.S+k  [1]",
       "z^2         Qkk.S+k.S+k  [1]",
       "z^2     2*Pk.Fk.S+k.S+k  [2]",
       ""};
const char *term13s[13+2]={"",  //  9pairs+4single = 6(+3back)pairs+4single  
       "zz ", // [0..2] 1 ??? =gse1  (1..9 are bonds)
       "z+ ", // [0..1] 2
       "z- ", // [1..3] 3     =gse4
       "+z ", // [0..1] 4 = ~2
       "-z ", // [1..3] 5 = ~3
       "+- ", // [1..2] 6
       "-+ ", // [1..2] 7 = ~6
       "++ ", // [0   ] 8
       "-- ", // [1..4] 9  = gse2
       "z  ", // [1   ] 10     = gse6???  (10..13 are single site terms)
       "-  ", // [1..2] 11 = gse7???
       "z^2", // [1..2] 12 = gse8=zz???
       "+  ", // [0   ] 13
       ""};

#define MAX_out 200            // reduced 1000 to 200 Mar08

/* export for measure speed */
unsigned long
     nterms_out0=0, /* count output terms summed to zero (not stored) */
     nterms_out1=0, /* count output terms per thread (summed at end and stop output) */
     nterms_out2=0, /* overflow */
     nterms_in1=0,  /* count input  terms per thread (summed at end) */
     nterms_in2=0;  /* overflow */

#if DO_COLLECT
// gpg --gen-prime 1 31
#define HASH_PRIME 0x701F35F3   // not used @ moment



// work: indices will be sorted (not used in the calling routine), Joerg Sep05
//       same or permutated indices[2..5] will be stored as sum only once
//       indices pointing to configurations 1..NumberConfigurations
//       permutated indices dont have the same probability! that means:
//       its difficult to simplify (no easy symmetry)
//       but it looks like that every term appears twice (ToDo: simplify!)
//
// id: 0=gs_terms, 1=terms 2=expectations 3=excitation
//
void collect_term(int indices[6], double dummy, int id) 
{
  int  k; long i, j; // , nio=7;
  static double na_in=0, na_out=0; /* Apr08 long to dbl, no bond sum */
  static int  id_old=-1;         /* flag for new collection */
  static long counter=0;         /* buffer index for desciptors,multipliers */
  static long max_counter=0;     /* maximum buffer size */
  static int    descriptors[MAXNUM_HASHED_TERMS+1][6]; // 24 byte
  static double multipliers[MAXNUM_HASHED_TERMS+1];    //  8 byte
  // it is very slow to search a big table for an entry sequentialy
  // so we introduce a hash table of indizes to that entries
  //  and a table of indices to that table to have fast sequential access 
  //  to spare filled hash table (Joerg Sep05, xxzLC-14 5x faster)
  // number of hashed entries is in counter
  static long hash[MAXNUM_HASHED_TERMS+1]; // 8B hashed index to entries
  static int  hidx[MAXNUM_HASHED_TERMS+1]; // 4B index to hash entries
  static int hcol=0;   // counter for hash collisions, just as info
  static int hhit=0;   // counter for hash hits
  static int hnew=0;   // counter for new hash entries
  static int hval=-1;  // hash value
  // count number of terms, only verbose output
  static unsigned long
    numiterms[81+1], // input terms from obtain, 81 or 62 for id=0..2, ...
    numoterms[14*5]; // counting for each term13=1..13 for 0..4-coefficients
  long maxterms_out=0, maxterms_in=0;
  const int *map;
  map=((id==3)?map81:map62); // choose the right mapping


  if (nterms_in1==0 && nterms_in2==0 && id_old!=id) {  // new file
    for (i=0;i<=81; i++) numiterms[i]=0; // initialize counter
    for (i=0;i<14*5;i++) numoterms[i]=0; // initialize counter
  }

  nterms_in1++; if(!nterms_in1) nterms_in2++;
  if (hval<0) { hval=0; // used as flag, initialize hash once!
    for (i=0;i<MAXNUM_HASHED_TERMS+1;i++) { hidx[i]=hash[i]=0; }
  }

  if (counter){  // something stored into buffer
      /* save collected terms and reset buffer for next term */

    if (    descriptors[counter][0] < indices[0]  /* next configuration number? */
       // indices[1]: term
       || map[descriptors[counter][1]] != map[indices[1]]  /* next term? 2..62, Apr08 JS */
//     || descriptors[counter][1] != indices[1]  /* next term? 2..62, Apr08 JS */
       ||                       0  > indices[0] /* flush buffer condition */
      /* Sept 05: old was 0>i[0] but collect_gse needs 0>=i[0] */
//       ||    NumberConfigurations < indices[0]  # not id=3
       || counter >= MAXNUM_HASHED_TERMS) /* buffer full */
    {
        if (       MAXNUM_HASHED_TERMS == counter) {
           static long nflush=0;
           if (nflush<10) {
             printf("MAXNUM_HASHED_TERMS is to small, flush buffer\n");
             printf("Overflow happens for cfg=%3d term%02d %3d\n",
               indices[0],indices[1],map[indices[1]]);
             fflush(stdout);
           }
           if (++nflush==10)
            printf("excessive output is stopped after %d prints\n",MAX_out);
        }
	// first save information 
	for(j=1;j<=counter;j++){
	  
          if (!multipliers[j]) { nterms_out0++; continue; }  // summed to zero
          term_store(id,descriptors[j],multipliers[j]);
          nterms_out1++; if(!nterms_out1) nterms_out2++;
          for (i=0;descriptors[j][i+2] && i<4;i++);
          numoterms[ (map[descriptors[j][1]]) * 5 + i ]++;
	  // printf out information
	  if (Control.output > 1 && myid == 0 && nterms_out1 < MAX_out) {
	    printf("OUT: Cfg=%3d Term%02d    CCM= %3d %3d %3d %3d"
		   " mult=% 7.3f %s  [%d]\n",
		   descriptors[j][0],
               map[descriptors[j][1]],
		   descriptors[j][2], descriptors[j][3],
                   descriptors[j][4], descriptors[j][5], multipliers[j],
                          term13s[map[descriptors[j][1]]], (int)i);
            if (nterms_out1+1==MAX_out)
              printf("OUT: no output after MAX_out=%d terms\n", MAX_out);
          }

          hash[ hidx[j] ]=0; // reset hash entry
          hidx[j]=0; // not necessary
	  // now overwrite the old stored information 
	}
        for (i=1;i<=counter;i++) { hash[ hidx[i] ]=0; }
	counter=0;
      }
  }

  // ToDo: print OUT term description
  // i[2]>0 and i[3]=0 1st order coeff.
  // i[3]>0 and i[4]=0 2nd order coeff.
  // i[4]>0 and i[5]=0 3th order coeff.
  // i[5]>0            4th order coeff.
  if (Control.output > 2 && nterms_out1 < MAX_out)
    printf("IN:  Cfg=%3d Term%02d.%02d CCM= %3d %3d %3d %3d"
           " mult=% 7.3f %.23s\n",
       indices[0], map[indices[1]],
       indices[1],
       indices[2], indices[3],
       indices[4], indices[5], dummy,
       ((id==3)?term81s[indices[1]]:term62s[indices[1]]));
  if (Control.output > 2 && myid == 0 && nterms_out1+1 == MAX_out)
    printf("IN:  Output stopped\n");

  // the last config has lot of terms, try to reduce further by
  // reducing all possible permutations to one config by sorting (big first)

  // for id=3 (excitations):
  //   x[2]    <= excitation coefficients
  //   x[3..5] <= g-s ket-state coefficients (NO PARTICULAR ORDER)
  // if (id!=3)  // Joerg Jun08: sorting for Excitad states failes
  for(;;) { // bubble sort indices
    j=1;
    for(i=((id==3)?3:2); i<5; i++) // for term_exc sort 3..5 only
      if (indices[i]<indices[i+1]) {  // exchange two indices
        j=0; k=indices[i+1]; indices[i+1]=indices[i]; indices[i]=k; 
      }
    if (j) break;
  }

  // Oct05: new hash table search (fast)
  j=0;
  if (indices[0]>=0)
  {
    
    hval=(((((                          indices[5])
              *(NumberConfigurations+1)+indices[4])
              *(NumberConfigurations+1)+indices[3])
              *(NumberConfigurations+1)+indices[2])
              *(NumberConfigurations+1)+indices[0]);
    if (id==3)
    hval=(((((                                 indices[5])
              *(NumberExcitedConfigurations+1)+indices[4])
              *(NumberExcitedConfigurations+1)+indices[3])
              *(NumberExcitedConfigurations+1)+indices[2])
              *(NumberExcitedConfigurations+1)+indices[0]);
    // try up to 1024 times,
    for(i=0;;i++) {
      hval = hval + indices[2] + 1;  // avoid overflow to negativ values
      // abs(hval) will fail for INT32_MIN
      if (hval<0) hval=~hval; // INT32_MIN..-1 <-> INT32_MAX..0
      hval %= MAXNUM_HASHED_TERMS; // MAX should be prime
      j = hash[hval];
      if (j>counter || j<0 || hval<0) { 
         printf("Bug86757 j=%ld counter=%ld hash=%d\n",j,counter,hval);
         exit(1);
      }
      if (j==0) {      // empty entry
             hash[hval]=counter+1;  // set hash entry
             hidx[counter+1]=hval;  // set sequential index to entry
             j=counter+1;           // store to table
             hnew++; break;
      } else {
        if ( indices[0]==descriptors[j][0] &&
         map[indices[1]]==map[descriptors[j][1]] &&  // term
             indices[2]==descriptors[j][2] &&
             indices[3]==descriptors[j][3] &&
             indices[4]==descriptors[j][4] &&
             indices[5]==descriptors[j][5]
           ) {
            multipliers[j]+=dummy;  // hashed == searched
            hhit++; break;
         } else if (i>=1024) {
           static int hcol_out=1;
           hcol++; nterms_out1++;    // collision, write through
           if (!nterms_out1) nterms_out2++;
           { int i=0;
             for (i=0;descriptors[j][i+2] && i<4;i++);
             numoterms[ (map[indices[1]]) * 5 + i ]++;
           }
           term_store(id,indices,dummy);
           if (hcol_out==hcol) { // avoid excessiv output
             if (hcol_out<<1) hcol_out<<=1;  // *=2
             printf("%2d: collision hash[%d]=%ld lenhash=%ld=%.3f%%\n",
                     myid, hval, hash[hval], counter,
                     counter*1.0/MAXNUM_HASHED_TERMS);
           }
           // ToDo: create/access a 2nd level hash table using 4*hashsize
           //       or use a tree (num descriptorindizes, terms), count!
           j=0; break;
         }
      }
    }
  }
  
  if (0<=indices[0]) { // valid
    if (indices[1]<0 || indices[1]>((id==3)?81:62)) { 
      printf("Error collect.C L%d: term=%d\n",__LINE__,indices[1]); ERR(1);
    }
    numiterms[indices[1]]++; // for statistics only
  }

  if (0<=indices[0])
  if ( j>counter /* && indices[1] != 0 */ ) {
    counter++;  /* potential overflow! */
    if (counter>max_counter) max_counter=counter; /* just as info */
    descriptors[counter][0]=indices[0];
    descriptors[counter][1]=indices[1];
    descriptors[counter][2]=indices[2];
    descriptors[counter][3]=indices[3];
    descriptors[counter][4]=indices[4];
    descriptors[counter][5]=indices[5];
    multipliers[counter]=dummy;
  }

  if (indices[0]<0) { // End-of-Data-Set
    term_store(id,indices,dummy);

    if (Control.output > 1 && myid == 0 ) {  // STOPTERM=999999
      printf("OUT: Cfg=%3d Term%02d\n", indices[0], indices[1]);
    }
    if (id_old != id) { id_old=id; na_out=na_in=0; } // new file
    if (Control.output > -1) {
      /* ToDo: clock output here? */
      int t2, tmin, tmax, tsum;
      t2=seconds(); // differ for each thread (collect min,max)
      tmin=tmax=tsum=t2;
      maxterms_in =nterms_in1;  /* set default for single thread computation */
      maxterms_out=nterms_out1;
#define ShiftULong (1.*(((unsigned long)1)<<(sizeof(long)*4))\
                   *1.*(((unsigned long)1)<<(sizeof(long)*4)))
      na_out=nterms_out2*ShiftULong + nterms_out1;
      na_in =nterms_in2 *ShiftULong + nterms_in1;
#if (PARALLEL_FLAG> 0)
      int            tmp[81+1], i;   // array used for 14*5, 81+1, use max!
      unsigned long tmp1[81+1], tmp2[81+1];
      double          d1[81+1],   d2[81+1];
      /* SUM may overflow */
      for(i=0;i<81+1;i++) tmp1[i]=numiterms[i];   // input terms
      MPI_Allreduce(tmp1,tmp2,81+1,MPI_UNSIGNED_LONG,MPI_SUM,MPI_COMM_WORLD);
      for(i=0;i<81+1;  i++) numiterms[i]=tmp2[i];   // input terms

      for(i=0;i<14*5;i++) d1[i]=tmp1[i]=numoterms[i];  // output terms
      MPI_Allreduce(tmp1,tmp2,14*5,MPI_UNSIGNED_LONG,MPI_SUM,MPI_COMM_WORLD);
      MPI_Allreduce(  d1,  d2,14*5,MPI_DOUBLE,       MPI_SUM,MPI_COMM_WORLD);
      for(i=0;i<14*5;i++) numoterms[i]=tmp2[i];   // output terms
      if (!myid)
        for(i=0;i<14*5;i++)
          if (d2[i]!=1.*numoterms[i])
            printf("Integer Overflow sum numoterms[%2d][%2d]: %13.6e\n",
                    i/5+1, i%5, d2[i]);
      if (myid>0) for(i=0;i<81+1;i++) numiterms[i]=0;
      if (myid>0) for(i=0;i<14*5;i++) numoterms[i]=0;
      MPI_Allreduce(    &t2, &tmin,  1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
      MPI_Allreduce(    &t2, &tmax,  1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
      MPI_Allreduce(    &t2, &tsum,  1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
      MPI_Allreduce(  &hcol,   tmp,  1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
      hcol=tmp[0];
      // measure balance
      MPI_Allreduce( &nterms_in1,tmp2, 1, MPI_LONG, MPI_MAX, MPI_COMM_WORLD);
      maxterms_in=tmp2[0];
      MPI_Allreduce(&nterms_out1,tmp2, 1, MPI_LONG, MPI_MAX, MPI_COMM_WORLD);
      maxterms_out=tmp2[0];
      // sum
      d1[0]=nterms_out2*ShiftULong+nterms_out1;
      d1[1]=na_in;
      d1[2]=na_out;
      tmp1[0]=nterms_out0;
      tmp1[1]=nterms_in1;
      tmp1[2]=nterms_out1;
      tmp1[3]=nterms_in2;
      tmp1[4]=nterms_out2;
      MPI_Allreduce(tmp1,tmp2,5,MPI_UNSIGNED_LONG,MPI_SUM,MPI_COMM_WORLD);
      MPI_Allreduce(  d1,  d2,3,MPI_DOUBLE,       MPI_SUM,MPI_COMM_WORLD);
      nterms_out0=tmp2[0];
      nterms_in1 =tmp2[1];
      nterms_out1=tmp2[2];
      nterms_in2 =tmp2[3];
      nterms_out2=tmp2[4];
      MPI_Allreduce(&max_counter,tmp2,1,MPI_UNSIGNED_LONG,MPI_MAX,MPI_COMM_WORLD);
      max_counter=tmp2[0];
      if (!myid)
        if (d2[0]!=1.*nterms_out1)
          printf("Integer Overflow sum nterms_out: %13.6e\n", d2[0]);
      na_in =d2[1];
      na_out=d2[2];
#endif
      /* output debugging infos */
      // output this as a statistic information
      if (myid==0) {
        double sumi=0, sumo=0;
        for (i=0;i<=((id==3)?81:62);i++){
          if (numiterms[i]) printf(" i_terms%02d.%02ld: %12lu  %s\n",
                        map[i], i, numiterms[i],
                        ((id==3)?term81s[i]:term62s[i]));
          sumi += numiterms[i];
        }
        for (i=0;i<14*5;i++){
          /* output max 3chars of term62s (as collected term13) */
          if (numoterms[i])
            printf(" o_terms%02d:    %12lu  %3.3s [%d] ToDo: f(angle)?\n",
              (int)i/5, numoterms[i], term13s[i/5], (int)(i%5));
          sumo += numoterms[i];
        }
        printf(" collect_term maxbuff= %8ld\n", max_counter);
        printf(" collect_term hash collisions= %6d\n", hcol);
        printf(" collect_term maximum unbalance in= %6.2f%% out= %6.2f%%\n",
                maxterms_in *100.*npe/nterms_in1 -100.,
                maxterms_out*100.*npe/nterms_out1-100.);
        printf(" collect_terms    double        in= %12.0f  out= %10.0f\n", na_in, na_out);
        printf(" collect_terms    long          in= %12lu  out= %10lu\n", nterms_in1, nterms_out1);
        printf("                  long overflow in= %12lu  out= %10lu\n", nterms_in2, nterms_out2);
        printf("                  summed to zero   %12s   out= %10lu\n", " ", nterms_out0);
        printf(" collect_term bond summed terms in= %12.0f  out= %10.0f\n", sumi,  sumo);
        printf(" collect_term storage memory per term: 6*int+dbl=%d (or compressed -25%%)\n",
                (int)(6*sizeof(int)+sizeof(double)));
        printf(" collect_term  t[min]= %6.2f .. %6.2f .. %6.2f  npe= %d id= %d\n\n",
                tmin/60., tsum/npe/60., tmax/60., npe, id);
      }
    }
    max_counter=0;
    nterms_out2=nterms_in2=0;
    nterms_out1=nterms_in1=0;
    hnew=hhit=hcol=0;   // reset number of collisions
    for (i=0;i<81+1;i++) numiterms[i]=0; // initialize counter
    for (i=0;i<14*5;i++) numoterms[i]=0; // initialize counter
  }
}
#else /* dont collect */
void collect_term(int indices[6],double dummy, int id) 
{
  int nio;

  term_store(id,indices,dummy);
	  
  if (Control.output > 1 && myid == 0 && nterms_out1 < MAX_out) {
     printf("OUT: Cfg=%3d Term%02d.%02d CCM= %3d %3d %3d %3d  mult=%10.7f\n",
     indices[0], map[indices[1]], indices[1],
     indices[2], indices[3],
     indices[4], indices[5], dummy);
     if (nterms_out1+1==MAX_out)
        printf("excessive output is stopped after %d terms\n", MAX_out);
  }
  if (indices[0]<0) {
    /* all terms collected, obtain finished for one bond */
#if (PARALLEL_FLAG> 0)
      long tmp;
      MPI_Allreduce(&nterms_out1,  &tmp,  1, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      nterms_out1=tmp;
#endif
    if (Control.output > 0 && myid == 0 ) {
      /* output debugging infos */
      printf(" collect_term  terms out = %8ld\n", nterms_out1);
      printf("                overflow = %8ld\n", nterms_out2);
      nterms_out2=nterms_out1=0;
    }
  }
  nterms_out1++; if(!nterms_out1) nterms_out2++;
  nterms_in1++;  if(!nterms_in1)  nterms_in2++; /* in++ for compability */
}
#endif

/* Jun08 Joerg: collect_term2    replaced by collect_term(...,id=2) */

/* Jun08 Joerg: collect_exc_term replaced by collect_term(...,id=3) */
